#include <math.h>
#include "nrutil.h"
#include "funcs.h"

void lubksb(double **a, int n, int indx[], double b[], double x[])
{
  int i,ll,j;
  double *bb,sum,*y;

  bb=dvector(1,n);
  y=dvector(1,n);

  for(i=1;i<=n;i++) bb[i]=b[i];

  for(i=1;i<=n;i++) {
    ll=indx[i];
    sum=bb[ll];
    bb[ll]=bb[i];
    for(j=1;j<i;j++) sum-=a[i][j]*y[j];
    y[i]=sum;
  }

  for(i=n;i>=1;i--) {
    sum=y[i];
    for(j=i+1;j<=n;j++) sum-=a[i][j]*x[j];
    x[i]=sum/a[i][i];
  }
  free_dvector(bb,1,n);
  free_dvector(y,1,n);
}



void ludcmp(double **a, int n, int indx[], double *d, double *khad)
{
  const double tiny=1.e-20;
  int i,j,k,imax;
  double sum,aamax,dum;

  *khad=1.0;
  for(i=1;i<=n;i++) {
    sum=0.0;
    for(j=1;j<=n;j++) sum+=a[i][j]*a[i][j];
    (*khad)*=sqrt(sum);
  }

  *d=1.0;
  for(j=1;j<=n;j++) {
    for(i=1;i<j;i++) {
      sum=a[i][j];
      for(k=1;k<i;k++) sum-=a[i][k]*a[k][j];
      a[i][j]=sum;
    }

    aamax=0.0;
    for(i=j;i<=n;i++) {
      sum=a[i][j];
      for(k=1;k<j;k++) sum-=a[i][k]*a[k][j];
      a[i][j]=sum;
      dum=fabs(sum);
      if(dum > aamax) {
        aamax=dum;
        imax=i;
      }
    }

    if(j != imax) {
      for(k=1;k<=n;k++) {
        dum=a[imax][k];
        a[imax][k]=a[j][k];
        a[j][k]=dum;
      }
      *d=-(*d);
    }

    indx[j]=imax;
    if(a[j][j]==0.0) a[j][j]=tiny;
    dum=1.0/a[j][j];
    for(i=j+1;i<=n;i++) a[i][j]*=dum;
    (*d)*=a[j][j];
  }
  *khad=fabs(*d)/(*khad);
}




void mrqcof(double x[], double y[], double sig[], int ndata, int ma,
	double a[], double **alpha, double beta[], double *chisq)
// Diese Funktion berechnet die Matrix ALPHA und den Vektor BETA gemaess
// Skriptum (4.18) sowie die gewichtete Fehlersumme chi**2.
// Quelle: C-NUMREC, S. 687.
{
	int i,j,k,pause;
	double ymod,g,dy,*dyda;

	dyda=dvector(1,ma);

	for (i=1;i<=ma;i++) {
		for (j=1;j<=i;j++) alpha[i][j]=0.0;
		beta[i]=0.0;
	}
	*chisq=0.0;
	for (k=1;k<=ndata;k++) {
		funcs(x[k],a,&ymod,dyda,ma);
		g=1.0/(sig[k]*sig[k]);
		dy=y[k]-ymod;
		for (i=1;i<=ma;i++) {
		  for (j=1;j<=i;j++) alpha[i][j] += g*dyda[i]*dyda[j];
		  beta[i] += g*dy*dyda[i];
		}
		*chisq += g*dy*dy;
	}
	for (i=2;i<=ma;i++)
	  for (j=1;j<i;j++) alpha[j][i]=alpha[i][j];

	free_dvector(dyda,1,ma);
}




int mrqmin(double x[], double y[], double sig[], int ndata, int ma,
           double a[], 
	   double **alpha, double **covar, double beta[], double *chisq, 
           double *ochisq, double *alamda)
// Diese Funktion fuehrt einen Iterationsschritt im Rahmen des Gauss-Newton-
// Marquardt-Verfahrens durch.
// Return-Wert der Funktion:   0   GNM-Schritt war erfolgreich,
//                             1   GNM-Schritt war nicht erfolgreich.
// Veraenderte Version aus C-NUMREC, S. 685f.
{
	int i,j,k,vmar,*indx,pause;
        double *atry,*da,*loes,*vektor,**normal,d,khad;

        atry=dvector(1,ma);
	da=dvector(1,ma);
	vektor=dvector(1,ma);
	loes=dvector(1,ma);
	indx=ivector(1,ma);
	normal=dmatrix(1,ma,1,ma);

	if (*alamda < 0.0) {
	  *alamda=0.0003;
	  mrqcof(x,y,sig,ndata,ma,a,alpha,beta,chisq);
          *ochisq=(*chisq);
	}
	for (i=1;i<=ma;i++) {    
	  for (j=1;j<=ma;j++)  normal[i][j]=alpha[i][j];
          normal[i][i]=alpha[i][i] * (1.0+(*alamda));
        }
    
        ludcmp(normal,ma,indx,&d,&khad);

        if(*alamda == 0.0) {
          for(j=1;j<=ma;j++) {
            for(i=1;i<=ma;i++)vektor[i]=0.0;
	    vektor[j]=1.0;
            lubksb(normal,ma,indx,vektor,loes);
            for(i=1;i<=ma;i++) covar[i][j]=loes[i];
          }
        }
        else {
          lubksb(normal,ma,indx,beta,da);
          for(j=1;j<=ma;j++) {
            atry[j]=a[j]+da[j];
            if (a[j]<0) atry[j]=-atry[j];
          }
          mrqcof(x,y,sig,ndata,ma,atry,normal,vektor,chisq);

	  if (*chisq <= *ochisq) {
		*alamda /= 5.0;
		*ochisq=(*chisq);
                vmar=0;
		for (i=1;i<=ma;i++) {
		  for (j=1;j<=ma;j++) alpha[i][j]=normal[i][j];
	          beta[i]=vektor[i];
                  a[i]=atry[i];
                }
           }
	   else {
		*alamda *= 5.0;
                vmar=1;
	   }
         }
         free_dvector(atry,1,ma);
         free_dvector(da,1,ma);
         free_dvector(vektor,1,ma);
         free_dvector(loes,1,ma);
         free_ivector(indx,1,ma);
         free_dmatrix(normal,1,ma,1,ma);

        return   vmar;
}
